# -*- coding: utf-8 -*-

"""
AWS Organizations related data model.
"""

import typing as T
import enum
import dataclasses
from datetime import datetime

from iterproxy import IterProxy


class BaseModel:
    pass


class ParentTypeEnum(str, enum.Enum):
    ROOT = "ROOT"
    ORGANIZATIONAL_UNIT = "ORGANIZATIONAL_UNIT"


@dataclasses.dataclass
class Parent(BaseModel):
    id: str = dataclasses.field()
    type: str = dataclasses.field()

    def is_root(self) -> bool:
        return self.type == ParentTypeEnum.ROOT.value

    def is_ou(self) -> bool:
        return self.type == ParentTypeEnum.ORGANIZATIONAL_UNIT.value


class ChildTypeEnum(str, enum.Enum):
    ACCOUNT = "ACCOUNT"
    ORGANIZATIONAL_UNIT = "ORGANIZATIONAL_UNIT"


@dataclasses.dataclass
class Child(BaseModel):
    id: str = dataclasses.field()
    type: str = dataclasses.field()

    def is_account(self) -> bool:
        return self.type == ChildTypeEnum.ACCOUNT.value

    def is_ou(self) -> bool:
        return self.type == ChildTypeEnum.ORGANIZATIONAL_UNIT.value


class AccountOrOrgUnitOrOrg:
    """
    Mixin class for Account, OrganizationUnit and Organization.

    They all have three common testing methods
    ``is_account()``, ``is_ou()``, and ``is_org()``.
    """
    def is_account(self) -> bool:
        return False

    def is_ou(self) -> bool:
        return False

    def is_org(self) -> bool:
        return False


class AccountStatusEnum(str, enum.Enum):
    ACTIVE = "ACTIVE"
    SUSPENDED = "SUSPENDED"
    PENDING_CLOSURE = "PENDING_CLOSURE"


class AccountJoinedMethodEnum(str, enum.Enum):
    INVITED = "INVITED"
    CREATED = "CREATED"


@dataclasses.dataclass
class Account(
    BaseModel,
    AccountOrOrgUnitOrOrg,
):
    """
    Represents an AWS Account.
    """
    id: T.Optional[str] = dataclasses.field(default=None)
    arn: T.Optional[str] = dataclasses.field(default=None)
    name: T.Optional[str] = dataclasses.field(default=None)
    email: T.Optional[str] = dataclasses.field(default=None)
    status: T.Optional[str] = dataclasses.field(default=None)
    joined_method: T.Optional[str] = dataclasses.field(default=None)
    joined_timestamp: T.Optional[datetime] = dataclasses.field(default=None)

    # relationship
    parent_obj: T.Union["Organization", "OrganizationalUnit"] = dataclasses.field(
        default=None
    )

    def is_account(self) -> bool:
        return True


@dataclasses.dataclass
class OrganizationalUnit(
    BaseModel,
    AccountOrOrgUnitOrOrg,
):
    """
    Represents an AWS Organization Unit.
    """
    id: T.Optional[str] = dataclasses.field(default=None)
    arn: T.Optional[str] = dataclasses.field(default=None)
    name: T.Optional[str] = dataclasses.field(default=None)

    # relationship
    parent_obj: T.Union["Organization", "OrganizationalUnit"] = dataclasses.field(
        default=None
    )
    org_units: T.List["OrganizationalUnit"] = dataclasses.field(default_factory=list)
    accounts: T.List[Account] = dataclasses.field(default_factory=list)

    def is_ou(self) -> bool:
        return True

    @property
    def org_units_names(self) -> T.List[str]:
        return [ou.name for ou in self.org_units]

    @property
    def accounts_names(self) -> T.List[str]:
        return [account.name for account in self.accounts]


@dataclasses.dataclass
class Organization(
    BaseModel,
    AccountOrOrgUnitOrOrg,
):
    """
    Represents an AWS Organization.
    """
    id: T.Optional[str] = dataclasses.field(default=None)
    arn: T.Optional[str] = dataclasses.field(default=None)
    feature_set: T.Optional[str] = dataclasses.field(default=None)
    master_account_arn: T.Optional[str] = dataclasses.field(default=None)
    master_account_id: T.Optional[str] = dataclasses.field(default=None)
    master_account_email: T.Optional[str] = dataclasses.field(default=None)
    available_policy_types: T.List[dict] = dataclasses.field(default_factory=list)

    # relationship
    parent_obj: None = dataclasses.field(default=None)
    org_units: T.List["OrganizationalUnit"] = dataclasses.field(default_factory=list)
    accounts: T.List[Account] = dataclasses.field(default_factory=list)

    @property
    def name(self) -> str:
        return f"Organization({self.id})"

    def is_org(self) -> bool:
        return True

    @property
    def org_units_names(self) -> T.List[str]:
        return [ou.name for ou in self.org_units]

    @property
    def accounts_names(self) -> T.List[str]:
        return [account.name for account in self.accounts]


# ------------------------------------------------------------------------------
# Iterproxy
# ------------------------------------------------------------------------------
class ParentIterproxy(IterProxy[Parent]):
    pass


class ChildIterproxy(IterProxy[Child]):
    pass


class AccountIterproxy(IterProxy[Account]):
    pass


class OrganizationUnitIterproxy(IterProxy[OrganizationalUnit]):
    pass



# -*- coding: utf-8 -*-

"""
Organizational Structure objective oriented interface.

Ref:

- Core concepts: https://docs.aws.amazon.com/whitepapers/latest/organizing-your-aws-environment/core-concepts.html
"""

import dataclasses
import typing as T

from anytree import NodeMixin, RenderTree

# from rich import print as rprint

from .better_boto import (
    ParentTypeEnum,
    Parent,
    ChildTypeEnum,
    Child,
    AccountStatusEnum,
    AccountJoinedMethodEnum,
    Account,
    OrganizationalUnit,
    Organization,
    AccountIterproxy,
    OrganizationUnitIterproxy,
    list_parents,
    list_children,
    get_root_id,
    list_organizational_units_for_parent,
    list_accounts_for_parent,
    describe_organization,
)

if T.TYPE_CHECKING:
    from boto_session_manager import BotoSesManager


class Node(NodeMixin):
    """
    A node on the organization structure Tree.

    :param id: the id of the object on the node
    :param name: human friendly name, the name of the object on the node
    :param obj: the object on the node could be one of
        Organization, OrganizationUnit, and Account
    """

    def __init__(
        self,
        id: str,
        name: str,
        type: str,
        obj: T.Union[Organization, OrganizationalUnit, Account],
        parent=None,
        children=None,
    ):
        self.id = id
        self.name = name
        self.type = type
        self.obj = obj
        self.parent = parent
        if children:
            self.children = children

    def __repr__(self) -> str:
        return f"{self.name} ({self.type} {self.id!r})"

    @property
    def path_key(self) -> str:
        return " | ".join([
            node.name
            for node in self.path
        ])

    def _iter_accounts(self, recursive: bool = True) -> T.Iterable[Account]:
        node: Node
        if recursive:
            for _, _, node in RenderTree(self):
                if node.obj.is_account():
                    yield node.obj
        else:
            for node in self.root.children:
                if node.obj.is_account():
                    yield node.obj

    def iter_accounts(self, recursive: bool = True) -> AccountIterproxy:
        return AccountIterproxy(self._iter_accounts(recursive=recursive))

    def _iter_org_units(self, recursive: bool = True) -> T.Iterable[Account]:
        node: Node
        if recursive:
            for _, _, node in RenderTree(self):
                if node.obj.is_ou():
                    yield node.obj
        else:
            for node in self.root.children:
                if node.obj.is_ou():
                    yield node.obj

    def iter_org_units(self, recursive: bool = True) -> OrganizationUnitIterproxy:
        return OrganizationUnitIterproxy(self._iter_org_units(recursive=recursive))


@dataclasses.dataclass
class OrgStructure:
    """
    Abstraction of the AWS Organization structure.

    It is a tree structure of Organization, OrganizationalUnit, and Account.

    API:

    - ``self.root`` is the root node of the tree.
    - ``self.visualize()`` can visualize the tree.
    - ``for ou in self.root.iter_org_units(recursive=True):`` can iterate all OU.
    - ``for acc in self.root.iter_org_accounts(recursive=True):`` can iterate all Accounts.
    - ``self.is_x_in_y()`` can test if an account / ou is in an ou or org.
    """
    root: Node = dataclasses.field()

    _mapper: T.Dict[str, Node] = dataclasses.field(init=False, default_factory=dict)

    def __post_init__(self):
        self._mapper[self.root.id] = self.root
        self._mapper[self.root.obj.id] = self.root
        node: Node
        for _, _, node in RenderTree(self.root):
            self._mapper[node.id] = node

    def visualize(self):
        """
        Visualize (print) the organization structure tree.
        """
        print(RenderTree(self.root))

    def to_csv(self, sep="\t") -> str:
        rows = [
            ("Type", "Path", "Id")
        ]
        node: Node
        for pre, fill, node in RenderTree(self.root):
            rows.append(
                (
                    node.type,
                    node.path_key,
                    node.id,
                )
            )
        return "\n".join([
            sep.join(row)
            for row in rows
        ])

    def get_node(self, id: str) -> Node:
        """
        Get a node by id.
        """
        return self._mapper[id]

    def _resolve_node(
        self,
        node_or_object_or_id: T.Union[
            Node, Organization, OrganizationalUnit, Account, str
        ],
    ) -> Node:
        if isinstance(node_or_object_or_id, str):
            return self._mapper[node_or_object_or_id]
        elif isinstance(node_or_object_or_id, Node):
            return node_or_object_or_id
        else:
            return self._mapper[node_or_object_or_id.id]

    def _is_x_in_y(
        self,
        node_or_object_or_id_x: T.Union[
            Node, Organization, OrganizationalUnit, Account, str
        ],
        node_or_object_or_id_y: T.Union[Node, Organization, OrganizationalUnit, str],
    ) -> bool:
        node_x = self._resolve_node(node_or_object_or_id_x)
        node_y = self._resolve_node(node_or_object_or_id_y)
        return node_y.id in {ou.id for ou in node_x.ancestors}

    def is_x_in_y(
        self,
        x: T.Union[Node, Organization, OrganizationalUnit, Account, str],
        y: T.Union[Node, Organization, OrganizationalUnit, Account, str],
    ) -> bool:
        """
        Test if an account / ou is in an ou or org.
        """
        return self._is_x_in_y(x, y)

    @classmethod
    def get_org_structure(cls, bsm: "BotoSesManager") -> "OrgStructure":
        """
        Get the root node of the organization structure tree.

        :param bsm: the boto session manager of the management AWS Account (Root)
            of this organization.
        """
        org = describe_organization(bsm=bsm)

        root_id = get_root_id(bsm=bsm, aws_account_id=bsm.aws_account_id)

        ROOT = Node(id=root_id, name="Root", type="ROOT", obj=org)

        def walk_through(root: Node):
            """
            depth first search to walk through the organization structure tree or
            organization unit.
            """
            for ou in list_organizational_units_for_parent(bsm=bsm, parent_id=root.id):
                ou.parent_obj = root.obj
                root.obj.org_units.append(ou)
                leaf = Node(
                    id=ou.id,
                    name=ou.name,
                    type="Org Unit",
                    obj=ou,
                    parent=root,
                )
                walk_through(leaf)

            for account in list_accounts_for_parent(bsm=bsm, parent_id=root.id):
                account.parent_obj = root.obj
                root.obj.accounts.append(account)
                leaf = Node(
                    id=account.id,
                    name=account.name,
                    type="Account",
                    obj=account,
                    parent=root,
                )

        walk_through(ROOT)

        # print(RenderTree(ROOT))

        # rprint(ROOT.obj.parent_obj)
        # rprint(ROOT.obj.org_units_names)
        # rprint(ROOT.obj.accounts_names)

        # rprint(ROOT.obj.org_units[0].name)
        # rprint(ROOT.obj.org_units[0].org_units_names)
        # rprint(ROOT.obj.org_units[0].accounts_names)

        # rprint(ROOT.obj.org_units[0].parent_obj.arn)

        return OrgStructure(root=ROOT)



# -*- coding: utf-8 -*-

"""
Organization related boto3 API enhancement.
"""

import typing as T

from .model import (
    Organization,
)

if T.TYPE_CHECKING:
    from boto_session_manager import BotoSesManager


def describe_organization(
    bsm: "BotoSesManager",
) -> Organization:
    response = bsm.organizations_client.describe_organization()
    return Organization(
        id=response["Organization"]["Id"],
        arn=response["Organization"]["Arn"],
        feature_set=response["Organization"]["FeatureSet"],
        master_account_arn=response["Organization"]["MasterAccountArn"],
        master_account_id=response["Organization"]["MasterAccountId"],
        master_account_email=response["Organization"]["MasterAccountEmail"],
        available_policy_types=response["Organization"]["AvailablePolicyTypes"],
    )



# -*- coding: utf-8 -*-

from .model import (
    ParentTypeEnum,
    Parent,
    ChildTypeEnum,
    Child,
    AccountStatusEnum,
    AccountJoinedMethodEnum,
    Account,
    OrganizationalUnit,
    Organization,

    ParentIterproxy,
    ChildIterproxy,
    AccountIterproxy,
    OrganizationUnitIterproxy,
)

from .org_unit import (
    list_parents,
    list_children,
    get_root_id,
    list_organizational_units_for_parent,
    list_accounts_for_parent,
)

from .org import (
    describe_organization,
)



# -*- coding: utf-8 -*-

"""
AWS Organizations SDK enhancement.
"""


from ._version import __version__

__short_description__ = "AWS Organizations SDK enhancement."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"

try:
    from . import better_boto
    from .org_struct import (
        Node,
        OrgStructure,
    )
except ImportError:
    pass



# -*- coding: utf-8 -*-

"""
Organizational unit related boto3 API enhancement.
"""

import typing as T

from .model import (
    Parent,
    Child,
    OrganizationalUnit,
    Account,
    ParentIterproxy,
    ChildIterproxy,
    OrganizationUnitIterproxy,
    AccountIterproxy,
)

if T.TYPE_CHECKING:
    from boto_session_manager import BotoSesManager


def _list_parents(
    bsm: "BotoSesManager",
    child_id: str,
    page_size: int = 20,
    max_results: int = 1000,
) -> T.Iterator[Parent]:
    paginator = bsm.organizations_client.get_paginator("list_parents")
    for response in paginator.paginate(
        ChildId=child_id,
        PaginationConfig=dict(
            PageSize=page_size,
            MaxResults=max_results,
        ),
    ):
        for dct in response.get("Parents", []):
            yield Parent(
                id=dct["Id"],
                type=dct["Type"],
            )


def list_parents(
    bsm: "BotoSesManager",
    child_id: str,
    page_size: int = 20,
    max_results: int = 1000,
) -> ParentIterproxy:
    return ParentIterproxy(
        _list_parents(
            bsm=bsm,
            child_id=child_id,
            page_size=page_size,
            max_results=max_results,
        )
    )


def _list_children(
    bsm: "BotoSesManager",
    parent_id: str,
    child_type: str,
    page_size: int = 20,
    max_results: int = 1000,
) -> T.Iterator[Child]:
    paginator = bsm.organizations_client.get_paginator("list_children")
    for response in paginator.paginate(
        ParentId=parent_id,
        ChildType=child_type,
        PaginationConfig=dict(
            PageSize=page_size,
            MaxResults=max_results,
        ),
    ):
        for dct in response.get("Children", []):
            yield Child(
                id=dct["Id"],
                type=dct["Type"],
            )


def list_children(
    bsm: "BotoSesManager",
    parent_id: str,
    child_type: str,
    page_size: int = 20,
    max_results: int = 1000,
) -> ChildIterproxy:
    return ChildIterproxy(
        _list_children(
            bsm=bsm,
            parent_id=parent_id,
            child_type=child_type,
            page_size=page_size,
            max_results=max_results,
        )
    )


def get_root_id(
    bsm: "BotoSesManager",
    aws_account_id: str,
) -> str:
    for parent in list_parents(bsm=bsm, child_id=aws_account_id):
        if parent.is_root():
            return parent.id
    raise ValueError("Could not find root id")


def _list_organizational_units_for_parent(
    bsm: "BotoSesManager",
    parent_id: str,
    page_size: int = 20,
    max_results: int = 1000,
) -> T.Iterator[OrganizationalUnit]:
    paginator = bsm.organizations_client.get_paginator(
        "list_organizational_units_for_parent"
    )
    for response in paginator.paginate(
        ParentId=parent_id,
        PaginationConfig=dict(
            PageSize=page_size,
            MaxResults=max_results,
        ),
    ):
        for dct in response.get("OrganizationalUnits", []):
            yield OrganizationalUnit(
                id=dct["Id"],
                arn=dct["Arn"],
                name=dct["Name"],
            )


def list_organizational_units_for_parent(
    bsm: "BotoSesManager",
    parent_id: str,
    page_size: int = 20,
    max_results: int = 1000,
) -> OrganizationUnitIterproxy:
    return OrganizationUnitIterproxy(
        _list_organizational_units_for_parent(
            bsm=bsm,
            parent_id=parent_id,
            page_size=page_size,
            max_results=max_results,
        )
    )


def _list_accounts_for_parent(
    bsm: "BotoSesManager",
    parent_id: str,
    page_size: int = 20,
    max_results: int = 1000,
) -> T.Iterator[Account]:
    paginator = bsm.organizations_client.get_paginator(
        "list_accounts_for_parent"
    )
    for response in paginator.paginate(
        ParentId=parent_id,
        PaginationConfig=dict(
            PageSize=page_size,
            MaxResults=max_results,
        ),
    ):
        for dct in response.get("Accounts", []):
            yield Account(
                id=dct["Id"],
                arn=dct["Arn"],
                name=dct["Name"],
                email=dct["Email"],
                status=dct["Status"],
                joined_method=dct["JoinedMethod"],
                joined_timestamp=dct["JoinedTimestamp"],
            )


def list_accounts_for_parent(
    bsm: "BotoSesManager",
    parent_id: str,
    page_size: int = 20,
    max_results: int = 1000,
) -> AccountIterproxy:
    return AccountIterproxy(
        _list_accounts_for_parent(
            bsm=bsm,
            parent_id=parent_id,
            page_size=page_size,
            max_results=max_results,
        )
    )
